var watch = require("watch"),
		less = require("less"),
		colors = require("colors"),
		growl = require("./growl"),
		sys = require("sys"),
		path = require("path"),
		util = require("util"),
		fs = require("fs");

// String helpers
String.prototype.__defineGetter__('ext',function() { return this.split(".")[this.split(".").length-1]; });
String.prototype.replaceExt = function(value) {	return this.replace("."+this.ext,"."+value); };

// Statics
var fileDelimiter = "//::file::";
var notify = {

	added : function(message)
	{
		sys.puts("\n"+message.green);
		sys.puts("Recompiling CSS.");
		notify.growl(message, "Added");
	},
	removed : function(message)
	{
		sys.puts("\n"+message.red);
		sys.puts("Recompiling CSS.");
		notify.growl(message, "Removed");
	},
	changed : function(message)
	{
		sys.puts("\n"+message.yellow);
		sys.puts("Recompiling CSS.");
		notify.growl(message, "Updated");
	},
	error : function(message) {
		sys.puts("\n"+message.red);
		notify.growl(message, "Error!");
	},
	success : function(message) {
		sys.puts("\n"+message.green);
		notify.growl(message, "Success");
	},
	growl : function(message, title) {
		growl.notify(message, {title:title,image:__dirname+"/icon.icns"});
	}
};




// Get less file content
function readContent(file)
{
	// Mark beginning of file for error reporting
	var content = fileDelimiter + file + "\n";

	// Add file content to buffer
	content += fs.readFileSync(file,["utf-8"]);

	return content;
}


// Do the deed
function compileLess(target, lessFiles)
{
	var buffer = "";
	lessFiles.forEach(function(file){
		buffer += file.content;
	});

	var parser = new(less.Parser);

	try {

		parser.parse(buffer,function(err, tree)
		{
			if (err) {

				outputLessError(err, buffer);

			} else {

				var css = tree.toCSS();

				fs.writeFile(target, css, function(err)
				{
					if (err) notify.error("Could not write CSS file.\n"+err.message);
				});

				notify.success("Less compiled successfully.");
			}
		});

	} catch (err) {

		outputLessError(err, buffer);

	}
}


function outputLessError(err, buffer) {

	var lines = buffer.split("\n");
	var line = "";
	var linesLeftToSearch = err.line;
	var file = "";
	var lineInFile = 0;
	var adjacentLinesToShow = 2;
	var i;
	var message = "";

	while (linesLeftToSearch) {
		line = lines[linesLeftToSearch];
		if (line.indexOf(fileDelimiter) > -1 ) {
			file = line.split(fileDelimiter)[1];
			lineInFile = err.line - linesLeftToSearch - 1;
			break;
		}
		linesLeftToSearch--;
	}

	message = "Error: " + err.message +
			"\nFile: " + file +
			"\nLine: " + lineInFile + ", Col: " + err.column +
			"\n\n";

	for (i=0; i<adjacentLinesToShow*2+1; i++)
	{
		message += lines[i+err.line-adjacentLinesToShow] + "\n";
		if (i == adjacentLinesToShow-1) {
			var c = 0;
			while(c < err.column) {
				message += " ";
				c++;
			}
			message += "^\n";
		}
	}

	notify.error(message);
}



function watchFolders(cwd,folders,onchange,oncreate,onremove)
{
	if (!folders) return;

	folders.forEach(function(folder,i,a)
	{
		watch.createMonitor(cwd+folder, function(monitor)
		{
			monitor.files[ cwd + folder ];
			monitor.on("changed", onchange);
			monitor.on("created", oncreate);
			monitor.on("removed", onremove);
		});
	});
}


function findFilesWithExtInFolders(ext, cwd, folders, callback)
{
	if (!folders) return;

	var filesFound = [];
	var foldersSearched = 0;

	folders.forEach(function(folder,i,a)
	{
		watch.walk(cwd+folder, function(err,files){

			if (err) {
				callback(err);
			}

			for ( var file in files ) {
				if (ext == file.ext)
					filesFound.push(file);
			}

			foldersSearched++;

			if (foldersSearched == folders.length)
				callback(null, filesFound);
		});
	});
}


function buildLessFile(file, baseViewPath, callback)
{
	var lessFile = "";
	var viewId = "";
	var lessContent = "";

	function ucase(){
		return arguments[0].toUpperCase();
	}

	// Make a corresponding less file path
	lessFile = file.replaceExt("less");

	// If a less doesn't exist, create one
	try
	{
		fs.statSync(lessFile);
	}
	catch (e)
	{
		// Using some replace/regex to build a view Id for the CSS.

		// get entire path up to the base view path
		var path = file.substr(0,file.indexOf(baseViewPath)+baseViewPath.length);

		// remove path
		viewId = file.replace(path,"").

			// remove the extension
				replaceExt("").

			// convert path to uppercase
				replace(/\b[a-z]/g, ucase).

			// remove slashes
				replace(/[/]/g, "");

		// Now build a less container with the ID
		lessContent = '// Generated with Lessup' +
				'\n\n#' + viewId +
				'\n{\n' +
				'\t// LessCSS goes here.' +
				'\n}';

		fs.writeFile(lessFile,lessContent,function(err){
			if (err) throw err;
			return callback(lessFile,lessContent);
		});
	}
}







module.exports.Watcher = function(options)
{

	this.compileTo = "/styles.css";
	this.cwd = null;
	this.buffer = "";
	this.viewExt = "html";
	this.lessFolders = ["/"];
	this.viewFolders = []; // leave empty
	this.lessFiles = [];
	this.viewFiles = [];


	/* Constructor */
	this._create = function() {

		var self = this;

		for ( var option in options ) {
			if (this.hasOwnProperty(option))
			{
				this[option] = options[option];
			}
		}

		if (this.cwd == null) throw "Lessup needs a reference to the app’s current working directory.";

		// Merge viewFolders into lessFolders to import any less files that get created
		this.viewFolders.forEach(function(folder){
			if (self.lessFolders.indexOf(folder)<0)
				self.lessFolders.push(folder);
		});

		var lessFilesAdded = [];

		watchFolders(

				this.cwd,
				this.lessFolders,

			// on change
				function (f, curr, prev)
				{
					switch (f.ext)
					{
						case self.viewExt:
							break;

						case "less":
							notify.changed("Less ("+f.replace(self.cwd,'')+") was changed.");
							self.lessFiles.update(f);
							compileLess(self.cwd+self.compileTo, self.lessFiles);
							break;
					}
				},

			// on create
				function (f, curr, prev)
				{
					switch (f.ext)
					{
						case self.viewExt:
							try {
								fs.statSync(f.replaceExt("less"));
								notify.added("View ("+f.replace(self.cwd,'')+") was added. Less file exists.");
							} catch (e) {
								lessFilesAdded.push(f.replaceExt("less"));
								notify.added("View ("+f.replace(self.cwd,'')+") was added. Adding Less file.");
								buildLessFile(f, self.viewFolders[0]/*TODO: fix self*/, function(file,content){
									self.lessFiles.add(file,content);
								});
							}
							break;

						case "less":
							var idx = lessFilesAdded.indexOf(f);
							if ( idx > -1 ) {
								lessFilesAdded = lessFilesAdded.slice(idx,1);
								break;
							}
							notify.added("Less ("+f.replace(self.cwd,'')+") was added.");
							self.lessFiles.add(f);
							compileLess(self.cwd+self.compileTo, self.lessFiles);
							break;
					}
				},

			// on remove
				function (f, stat)
				{
					switch (f.ext)
					{
						case self.viewExt:
							notify.removed("View ("+f.replace(self.cwd,'')+") was removed. Leaving Less file in place.");
							break;

						case "less":
							notify.removed("Less ("+f.replace(self.cwd,'')+") was removed.");
							self.lessFiles.remove(f);
							compileLess(self.cwd+self.compileTo, self.lessFiles);
							break;
					}
				}
				);


		// If view folder is defined, get all views
		if (this.viewFolders.length)
			findFilesWithExtInFolders(this.viewExt,this.cwd,this.viewFolders,function(err,files){

				if (!files) return;

				files.forEach(function(file){
					buildLessFile(file, self.viewFolders[0]/*TODO: fix this*/, function(file,content){
						self.lessFiles.add(file,content);
					});
				});
			});

		// Get all less files, then compile
		findFilesWithExtInFolders("less",this.cwd,this.lessFolders,function(err,files){

			if (!files) return;

			files.forEach(function(file){
				self.lessFiles.add(file);
			});

			compileLess(self.cwd+self.compileTo, self.lessFiles);
		});

	};

	// Extend lessFiles with some methods
	this.lessFiles.add = function(file,content) {
		this.push({
			file: file,
			content: content ? content : readContent(file)
		});
	};
	this.lessFiles.find = function(file) {
		return this.filter(function(f){
			if (f.file == file)
				return true;
		})[0];
	};
	this.lessFiles.update = function(file) {
		var f = this.find(file);

		if (f)
			f.content = readContent(file);
		else
			notify.error("Couldn't find file to update.");
	};
	this.lessFiles.remove = function(file) {
		var f = this.find(file);

		if (f)
			this.splice(this.indexOf(f),1);
		else
			notify.error("Couldn't find file to remove.");
	};

	this._create();

};